/*
*  File: GLTypes.h
*  Author: Oliver Kreylos <kreylos@cs>
*  Purpose: Several data types used with the GL implementation.
*  Version: $Id: GLTypes.h,v 1.2 2001/06/13 11:29:31 grosa Exp $
*/

#ifndef GLTYPES_INCLUDED
#define GLTYPES_INCLUDED

#include <math.h>
#include <string.h>


struct GLstate;


//ROBERT added structure containing s and t values
struct GLtext // A strucure to hold texture values
{
  public:

  GLfloat s,t;

  //constuctors
  GLtext(){};

  GLtext(GLfloat sVal, GLfloat tVal);

};

struct GLcolor // A structure to hold color values in RGB format
	{
	public:
	/* Elements: */
	GLfloat r, g, b;
	/* Constructors and destructors: */
	GLcolor(void) // Dummy constructor
		{
		};
	GLcolor(GLfloat sR, GLfloat sG, GLfloat sB); // Constructor without clamping
	/* Methods: */
	GLcolor& clamp(void); // Clamps the color values to the interval [0, 1]
	GLcolor& operator+=(const GLcolor& other);
	/* Friend functions: */
	friend GLcolor operator+(const GLcolor& c1,const GLcolor& c2);
	friend GLcolor operator*(GLfloat s,const GLcolor& c1);
	friend GLcolor operator*(const GLcolor& c1,GLfloat s);
	friend GLcolor operator*(const GLcolor& c1,const GLcolor& c2);
	};

struct GLvector // A structure to hold points and vectors in homogenuous coordinates
	{
	public:
	/* Elements: */
	GLfloat c[4]; // Homogenuous coordinates
	/* Constructors and destructors: */
	GLvector(void) // Dummy constructor
		{
		};
	GLvector(GLfloat sX, GLfloat sY, GLfloat sZ, GLfloat sW = 1.0) // Elementwise, constructs a point by default
		{
		c[0]=sX;
		c[1]=sY;
		c[2]=sZ;
		c[3]=sW;
		};
	/* Methods: */
	operator const GLfloat*(void) const // Converts vector to array of floats
		{
		return c;
		};
	operator GLfloat*(void) // Ditto
		{
		return c;
		};
	const GLfloat& operator()(int i) const // Accesses a vector element
		{
		return c[i];
		};
	GLfloat& operator()(int i) // Ditto
		{
		return c[i];
		};

	GLfloat length(void) const
	{
		return (GLfloat)(sqrt(c[0]*c[0]+c[1]*c[1]+c[2]*c[2]));
	};
	GLvector& homogenize(void)
		{
		if(c[3]!=1.0)
			{
			c[0]/=c[3];
			c[1]/=c[3];
			c[2]/=c[3];
			c[3]=1.0;
			}
		return *this;
		};
	GLvector& normalize(void)
		{
		GLfloat length=(GLfloat)(sqrt(c[0]*c[0]+c[1]*c[1]+c[2]*c[2]));

		c[0]/=length;
		c[1]/=length;
		c[2]/=length;
		c[3]=0.0;
		return *this;
		};
	GLvector operator+(const GLvector other) const
		{
		return GLvector(c[0]+other.c[0],c[1]+other.c[1],c[2]+other.c[2],c[3]+other.c[3]);
		};
	GLvector operator-(const GLvector other) const
		{
		return GLvector(c[0]-other.c[0],c[1]-other.c[1],c[2]-other.c[2],c[3]-other.c[3]);
		};
	GLvector operator*(GLfloat s)
		{
		return GLvector(c[0]*s,c[1]*s,c[2]*s,c[3]*s);
		};
	GLfloat operator*(const GLvector& other) const // Calculates the dot product of two vectors
		{
		return c[0]*other.c[0]+c[1]*other.c[1]+c[2]*other.c[2]+c[3]*other.c[3];
		};
	GLvector cross(const GLvector& other) const
		{
		return GLvector(c[1]*other.c[2]-c[2]*other.c[1],c[2]*other.c[0]-c[0]*other.c[2],c[0]*other.c[1]-c[1]*other.c[0],0.0);
		};
	};

struct GLmatrix // A structure to hold transformation matrices
	{
	public:
	/* Elements: */
	GLfloat m[4][4]; // Matrix components in row-major format
	/* Static matrix creation functions: */
	static const GLmatrix identity; // The identity matrix
	static GLmatrix translate(GLfloat tx, GLfloat ty, GLfloat tz); // Returns a translation matrix
	static GLmatrix scale(GLfloat sx, GLfloat sy, GLfloat sz); // Returns a scaling matrix
	static GLmatrix rotate(GLfloat angle, GLfloat rx, GLfloat ry, GLfloat rz); // Returns a rotation matrix
	static GLmatrix ortho(GLfloat left, GLfloat right, GLfloat bottom, GLfloat top, GLfloat nnear, GLfloat ffar); // Returns an orthogonal projection
	static GLmatrix frustum(GLfloat left, GLfloat right, GLfloat bottom, GLfloat top, GLfloat nnear, GLfloat ffar); // Returns a perspective projection
	/* Constructors and destructors: */
	GLmatrix(void) // Dummy constructor
		{
		};
	GLmatrix(GLfloat sDiagonal); // Creates a zero or identity matrix
	GLmatrix(const GLfloat sM[16]); // Loads the matrix from a column-major array
	/* Methods: */
	const GLfloat& operator()(int i,int j) const // Accesses row i, column j
		{
		return m[i][j];
		};
	GLfloat& operator()(int i,int j) // Ditto
		{
		return m[i][j];
		};
	GLmatrix& operator*=(const GLmatrix& other); // Multiplies the matrix from the right with another matrix
	GLmatrix& multiplyLeft(const GLmatrix& other); // Multiplies the matrix from the left with another matrix
	GLvector operator*(const GLvector& v) const; // Multiplies the matrix with a vector
	GLvector transposedMultiply(const GLvector& v) const; // Multiplies the matrix with a vector from the left
	GLdouble subDeterminant(int i, int j) const; // Calculates the determinant of a 3x3 submatrix - used for matrix inversion
	GLmatrix calcInverse(void) const; // Returns the inverse of this matrix
	};


// Structures used in WOGL.cpp, WOGL_incomplete.cpp
// Some structures to make life a little easier: //


//ROBERT added GLIP_ST
enum InterpolatedValues // Enumerated type for which values have to be interpolated
	{
	GLIP_NONE=0x0,GLIP_X=0x1,GLIP_Y=0x2,GLIP_Z=0x4,GLIP_RGB=0x8,GLIP_ST=0x10
	};


//ROBERT added GLtext to vertex information
struct GLprimitiveVertex // A structure to hold polygon vertices
	{
	GLcolor color; // The vertex color
	GLvector coord; // The vertex coordinates
	GLtext texture; // Texture coordinates
	};


//ROBERT added GLIP_ST to interpolation also added s,t,ds,dt
struct GLinterpolation // A structure holding values that are interpolated across primitives
{
	public:
	/* Elements: */
	GLfloat x, dx; // Value and increment for x
	GLfloat y, dy; // Value and increment for y
	GLfloat z, dz; // Value and increment for z
	GLfloat r, dr; // Value and increment for red
	GLfloat g, dg; // Value and increment for green
	GLfloat b, db; // Value and increment for blue
        GLfloat s, ds; // Value and increment for s
        GLfloat t, dt; //Value and increment for t
	/* Constructors and destructors: */
	GLinterpolation(void)
  {
	};
	GLinterpolation(const GLprimitiveVertex& pv)
		:x(pv.coord(0)), y(pv.coord(1)), z(pv.coord(2)),
		 r(pv.color.r), g(pv.color.g), b(pv.color.b),
		 s(pv.texture.s), t(pv.texture.t)
	{
	};
	GLinterpolation(GLint whichValues, 
                  const GLinterpolation& start, 
                  const GLinterpolation& end, 
                  GLfloat steps, GLfloat firstStep) 
	{
		if(whichValues & GLIP_X)
		{
			dx = (end.x - start.x) / steps;
			x = firstStep * dx + start.x;
		}
		if(whichValues & GLIP_Y)
		{
			dy = (end.y - start.y) / steps;
			y = firstStep * dy + start.y;
		}
		if(whichValues & GLIP_Z)
		{
			dz = (end.z - start.z) / steps;
			z = firstStep * dz + start.z;
		}
		if(whichValues & GLIP_RGB)
		{
			dr = (end.r - start.r) / steps;
			dg = (end.g - start.g) / steps;
			db = (end.b - start.b) / steps;
			r = firstStep * dr + start.r;
			g = firstStep * dg + start.g;
			b = firstStep * db + start.b;
		}
		if(whichValues & GLIP_ST)
		{
		  ds = (end.s - start.s) / steps;
		  dt = (end.t - start.t) / steps;
		  s = firstStep * ds + start.s;
		  t = firstStep * dt + start.t;
		}
	};
	/* Methods: */
	GLinterpolation& setColor(const GLcolor& color)
	{
		r = color.r;
		g = color.g;
		b = color.b;
		return *this;
	};
	GLinterpolation& init(GLint whichValues, const GLinterpolation& start, const GLinterpolation& end, GLfloat steps, GLfloat firstStep) 
	{
		if(whichValues & GLIP_X)
		{
			dx = (end.x - start.x) / steps;
			x = firstStep * dx + start.x;
		}
		if(whichValues & GLIP_Y)
		{
			dy = (end.y - start.y) / steps;
			y = firstStep * dy + start.y;
		}
		if(whichValues & GLIP_Z)
		{
			dz = (end.z - start.z) / steps;
			z = firstStep * dz + start.z;
		}
		if(whichValues & GLIP_RGB)
		{
			dr = (end.r - start.r) / steps;
			r = firstStep * dr + start.r;
			dg = (end.g - start.g) / steps;
			g = firstStep * dg + start.g;
			db = (end.b - start.b) / steps;
			b = firstStep * db + start.b;
		}
		if(whichValues & GLIP_ST)
		{
		  ds = (end.s - start.s) / steps;
		  dt = (end.t - start.t) / steps;
		  s = firstStep * ds + start.s;
		  t = firstStep * dt + start.t;
		}
		return *this;
	};
	GLinterpolation& increment(GLint whichValues)
		{
		if(whichValues & GLIP_X)
			x += dx;
		if(whichValues & GLIP_Y)
			y += dy;
		if(whichValues & GLIP_Z)
			z += dz;
		if(whichValues & GLIP_RGB)
			{
			r += dr;
			g += dg;
			b += db;
			}
		if(whichValues & GLIP_ST)
		{
		  s += ds;
		  t += dt;
		}
		return *this;
		};
};

// *****************
// * GLlightSource *
// *****************

struct GLlightSource // A structure holding all information about a light source
	{
	public:
	/* Elements: */
	bool enabled; // Flag whether the light source is enabled (on)

	GLvector position; // The light source position in eye coordinates if position[3] != 0.0,
                     // or the light source direction if position[3] == 0.0 (far point).

	GLvector spotDirection; // The preferred direction of a spot light source in eye coordinates
	GLfloat spotExponent; // The dropoff coefficient of a spot light source
	GLfloat spotCutoff; // The cutoff angle of a spot light source
	GLdouble spotCutoffCos; // The cosine of the cutoff angle

	GLfloat constantAttenuation; // The constant attenuation coefficient
	GLfloat linnearAttenuation; // The linnear attenuation coefficient
	GLfloat quadraticAttenuation; // The quadratic attenuation coefficient

	GLcolor ambient; // The ambient light source color
	GLcolor diffuse; // The diffuse light source color
	GLcolor specular; // The specular light source color
	};

struct GLviewport // A structure holding the coefficients of the viewport transformation
	{
	public:
	/* Elements: */
	int x, y, width, height; // The size and position of the viewport
	GLfloat* zBuffer; // The z-buffer associated with the viewport
	GLfloat xFact, xAdd; // The x scaling factor and the x offset
	GLfloat yFact, yAdd; // The y scaling factor and the y offset
	};

#define glNumLightSources 8
#define glModelviewStackDepth 32
#define glProjectionStackDepth 2

#define WOGL_UNSPECIFIED -1
#define WOGL_ORTHO 0
#define WOGL_PERSPECTIVE 1

struct GLstate // A structure to hold the current GL state
	{
	/* Elements: */
	public:
	GLcolor clearColor; // The color to clear the window
	GLfloat clearDepth; // The depth to clear the z buffer

	/* The current color and material: */
	GLcolor color; // The current color
	GLcolor ambient; // The current ambient color
	GLcolor diffuse; // The current diffuse color
	GLcolor specular; // The current specular color
	GLfloat shininess; // The current shininess
	GLcolor emission; // The current emissive color

	/* Texture variables ROBERT*/
	GLtext textCoordinates; // The current texture coordinates
	GLvoid *textMap; // Contains pixel values for texture map
	GLubyte textColorMax; //Max color value in texture map
 
	bool normalizeNormals; // The current automatic normal vector normalization flag
	GLvector normal; // The current normal vector in eye coordinates
	bool colorMaterialEnabled; // The current material mode enable flag
	GLenum colorMaterial; // The current material mode setting
	bool lightingEnabled; // The current lighting enable flag
	GLlightSource lights[glNumLightSources]; // The eight light sources
	GLenum polygonMode; // The current polygon drawing mode
	GLenum matrixMode; // The current matrix mode
	GLmatrix modelView[glModelviewStackDepth]; // The current modelview matrix stack
	GLmatrix normalView[glModelviewStackDepth]; // The inverse of the current modelview matrix
	int modelViewTop; // The index of the top of the modelview matrix stack
	GLmatrix projection[glProjectionStackDepth]; // The current projection matrix
	int projectionTop; // The index of the top of the projection matrix stack

  GLint viewMode;     // MJK, can be WOGL_ORTHO (glOrtho) or WOGL_PERSPECTIVE (glFrustum)
  
  GLviewport viewport; // The current viewport
	bool depthTestEnabled; // The current depth test enable flag
	bool goraudShadingEnabled; // The current Goraud shading enable flag
	bool inPrimitiveBracket; // A flag to indicate whether GL is inside a begin/end pair
	GLenum primitiveType; // The type of primitive, when inside a begin/end bracket
	bool stripTriangleEven; // Flag if the next strip triangle to draw has to be reversed in vertex order
  bool lineLoopFirstVertex; // flag if this is the first vertex in a line loop primitive
	bool texture2D; // Control for 2D Texture Mapping
	GLint textWidth; // Width of texture map
    GLint textHeight; //Height of texture map
	GLint textWrapS; //Wrap around S
	GLint textWrapT; // Wrap around T
	GLint textMinFilter; //Minification filter
	GLint textMagFilter; //Magnification filter
	GLint textEnvMode; // Texture environment mode

	/* Constructors and destructors: */
	GLstate(void)
		:clearColor(0.0f,0.0f,0.0f),clearDepth(1.0f),color(1.0f,1.0f,1.0f),ambient(0.2f,0.2f,0.2f),
		 diffuse(0.8f,0.8f,0.8f),specular(0.0f,0.0f,0.0f),shininess(0.0f),emission(0.0f,0.0f,0.0f),
		 normalizeNormals(false),normal(0.0f,0.0f,1.0f,0.0f),colorMaterialEnabled(false),
		 colorMaterial(GL_AMBIENT_AND_DIFFUSE),lightingEnabled(false),polygonMode(GL_FILL),
		 matrixMode(GL_MODELVIEW),modelViewTop(0),projectionTop(0),
		 depthTestEnabled(false),goraudShadingEnabled(true),inPrimitiveBracket(false),
		 primitiveType(GL_POLYGON),stripTriangleEven(true),lineLoopFirstVertex(false), 
		 texture2D(false), textWidth(0), textHeight(0), textWrapS(GL_CLAMP), textWrapT(GL_CLAMP),
		 textMinFilter(GL_NEAREST), textMagFilter(GL_NEAREST), textEnvMode(GL_MODULATE)
		 //ROBERT added texture initialization
		{
		int i;
		for(i=0;i<glNumLightSources;++i)
			{
			lights[i].enabled=i==0;
			lights[i].position=GLvector(0.0f,0.0f,1.0f,0.0f);
			lights[i].spotDirection=GLvector(0.0f,0.0f,-1.0f,0.0f);
			lights[i].spotExponent=0.0f;
			lights[i].spotCutoff=180.0f;
			lights[i].spotCutoffCos=-1.0f;
			lights[i].constantAttenuation=1.0f;
			lights[i].linnearAttenuation=0.0f;
			lights[i].quadraticAttenuation=0.0f;
			lights[i].ambient=GLcolor(0.0f,0.0f,0.0f);
			lights[i].diffuse=i==0?GLcolor(1.0f,1.0f,1.0f):GLcolor(0.0f,0.0f,0.0f);
			lights[i].specular=i==0?GLcolor(1.0f,1.0f,1.0f):GLcolor(0.0f,0.0f,0.0f);
			}
		
		modelView[0]=GLmatrix::identity;
		normalView[0]=GLmatrix::identity;
		projection[0]=GLmatrix::identity;

    viewMode = WOGL_UNSPECIFIED;
		};
	};



#endif
